/*
 * Copyright (c) 1995 Jim Lowe
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Mark Tinguely and Jim Lowe
 * 4. The name of the author may not be used to endorse or promote products 
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef METEORGRAB
/*#define METEOR_SINGLE /* Define if you want low frame rate/less pci activity*/
/*#define METEOR_PLANER /* Define if you want planer mode rather than packed */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/unistd.h>
#include <sys/fcntl.h>
#include <machine/ioctl_meteor.h>
#include "sized_types.h"
#include "vid_image.h"
#include "vid_code.h"
#include "meteor_grab.h"
#include <sys/types.h>
#include <sys/mman.h>
#include <tcl.h>
#include <tk.h>
#include <errno.h>

extern  Tcl_Interp *interp;

/*----------------------------------------------------------------------*/
int		video,			/* File number for video device */
		xmit_color;		/* Transmitting color (pal-colour:-) */
#ifdef METEOR_PLANER
uint8 *		yuv_data;		/* Pointer to data area */
#else
unsigned int *	yuv_data;		/* Pointer to data area */
#endif
struct meteor_geomet geom;		/* current geometry for meteor */


static int MeteorGrab_GetFrame(uint8 *y_data, int8 *uv_data)
{
register int row, column, size;
#ifdef METEOR_PLANER
register int8	*pyuv, *pue, *puo, *pve, *pvo;
#else
register unsigned int *pyuv;
register unsigned int data;
#endif
	 int tmp;

#ifdef	METEOR_SINGLE
	 int err;

	tmp = METEOR_CAP_SINGLE ;
        err = ioctl(video, METEORCAPTUR, &tmp);
#endif

	size = geom.rows * geom.columns;
	pyuv = yuv_data;

#ifdef METEOR_PLANER
	if(xmit_color) {
		pue = (int8 *) pyuv + size;
		tmp = size >> 2;
		pve = pue + tmp;
		puo = pve + tmp;
		pvo = puo + tmp;
	}

	for( row = 0; row < (geom.rows >> 1); row++) {
		for(column = 0; column < (geom.columns >> 1); column++) {
			*y_data++ = *pyuv++;	
			*y_data++ = *pyuv++;	
			if(xmit_color) {
				*uv_data++ = *pue++;
				*uv_data++ = *pve++;
			}
		}
		for(column = 0; column < (geom.columns >> 1); column++) {
			*y_data++ = *pyuv++;	
			*y_data++ = *pyuv++;	
			if(xmit_color) {
				*uv_data++ = *puo++;
				*uv_data++ = *pvo++;
			}
		}
	}
#else
	for (row = 0; row < size ; row+=2) {
		data = *pyuv++;
		*y_data++ = (data >>  8) & 0xff;
		*y_data++ = (data >> 24) & 0xff;
		if(xmit_color) {
			*uv_data++ =  data        & 0xff;
			*uv_data++ = (data >> 16) & 0xff;
		}
	}
#endif

    return (1);
}

/*
 * Set Meteor registers.
 * Usage: From Tcl:
 *	meteor get <variable>
 * 		returns the value of <variable>
 *	meteor set <variable> <value>
 *		
 */
#define NAME	0
#define	GETSET	1
#define	CMD	2
#define	VALUE	3
static char	*cmd_table[]  = {
	"FMT",
	"PORT",
	"HUE",
	"CHCV",
	NULL
};
#define	CMD_UNKNOWN	-1
#define	CMD_FMT		0
#define	CMD_PORT	1
#define	CMD_HUE		2
#define	CMD_CHCV	3
int
MeteorCmd(ClientData clientdata, Tcl_Interp *interp, int ac, char *av[])
{
int		set;
int		data;
static	char	buf[50];
int		error;
int		cmd;
int		i;

	error = TCL_OK;
	data = 0;
	cmd = CMD_UNKNOWN;
	interp->result = "0";

	if(ac < 3) {
		interp->result = "wrong # arguments";
		return TCL_ERROR;
	}
	set = 0;
	if(strcmp(av[GETSET], "set") == 0) {
		if(ac != 4) {
			interp->result = "wrong # arguments (set)";
			return TCL_ERROR;
		}
		set = 1;
		data = atoi(av[VALUE]);
	} else if (ac != 3) {	/* get */
		interp->result = "wrong # arguments (get)";
		return TCL_ERROR;
	}

	for(i=0; cmd_table[i] != NULL; i++) {
		if(strcmp(cmd_table[i], av[CMD]) == 0) {
			cmd = i;
			break;
		}
	}
	if(cmd == CMD_UNKNOWN) {
		sprintf(buf, "meteor: unknown command: %s", av[CMD]);
		interp->result = buf;
		return TCL_ERROR;
	}

	switch(cmd) {
	case	CMD_FMT:
		if(set) {
			if(data != 1 && data != 2 && data != 4 && data != 8) {
				sprintf(buf, "meteor: unknown fmt %d", data);
				error = TCL_ERROR;
			} else {
				data = data << 8;
        			if (ioctl(video, METEORSFMT, &data) < 0) {
					sprintf(buf, "meteor: METEORSFMT: %s",
						strerror(errno));
					error = TCL_ERROR;
				}
			}
		} else if (ioctl(video, METEORGFMT, &data) < 0) {
				sprintf(buf, "meteor: METEORGFMT: %s",
					strerror(errno));
				error = TCL_ERROR;
			} else
				sprintf(buf, "%d", data >> 8);
		break;
	case	CMD_PORT:
		if(set) {
			if(data != 1 && data != 2 && data != 4) {
				sprintf(buf, "meteor: unknown port %d", data);
				error = TCL_ERROR;
			} else {
				data = data << 12;
        			if (ioctl(video, METEORSINPUT, &data) < 0) {
					sprintf(buf, "meteor: METEORSINPUT: %s",
						strerror(errno));
					error = TCL_ERROR;
				}
			}
		} else if (ioctl(video, METEORGINPUT, &data) < 0) {
				sprintf(buf, "meteor: METEORGINPUT: %s",
					strerror(errno));
				error = TCL_ERROR;
			} else
				sprintf(buf, "%d", data >> 12);
		break;
	case	CMD_HUE:
		if(set) {
			char	c = (char)data;
			if(ioctl(video, METEORSHUE, &c) < 0) {
				sprintf(buf, "meteor: METEORSHUE: %s",
					strerror(errno));
				error = TCL_ERROR;
			}
		} else {
			char	c;
			if(ioctl(video, METEORGHUE, &c) < 0) {
				sprintf(buf, "meteor: METEORGHUE: %s",
					strerror(errno));
				error = TCL_ERROR;
			} else
				sprintf(buf, "%d", c);
		}
		break;
	case	CMD_CHCV:
		if(set) {
			u_char	c = (u_char)data;
			if(ioctl(video, METEORSCHCV, &c) < 0) {
				sprintf(buf, "meteor: METEORSCHCV: %s",
					strerror(errno));
				error = TCL_ERROR;
			}
		} else {
			u_char	c;
			if(ioctl(video, METEORGCHCV, &c) < 0) {
				sprintf(buf, "meteor: METEORGCHCV: %s",
					strerror(errno));
				error = TCL_ERROR;
			} else
				sprintf(buf, "%d", c);
		}
		break;
	}
	interp->result = buf;
	return error;
}

int
MeteorGrab_Probe(void)
{
int	temp;

	if ((video = open("/dev/meteor0", O_RDONLY)) < 0)
		return(0);

	/*
	 * Set up default parms.
	 */
	geom.rows = NTSC_HEIGHT * 2;	/* should be pal */
	geom.columns = NTSC_WIDTH * 2;	/* should be pal */
#ifdef METEOR_PLANER
	geom.oformat = METEOR_GEO_YUV_PLANER;
#else
	geom.oformat = METEOR_GEO_YUV_PACKED;
#endif
        geom.frames = 1;
	
	temp = ((geom.rows*geom.columns*2+4095)/4096)*4096;
#ifdef METEOR_PLANER
        yuv_data=(uint8 *) mmap((caddr_t)0, temp, PROT_READ, 0,
		video, (off_t)0);
	if (yuv_data == (uint8 *) -1) return (0);
#else
        yuv_data=(unsigned int *) mmap((caddr_t)0, temp, PROT_READ, 0,
		video, (off_t)0);
	if (yuv_data == (unsigned int *) -1) return (0);
#endif

	Tcl_CreateCommand(interp, "meteor", MeteorCmd, (ClientData)NULL,
			(Tcl_CmdDeleteProc *)NULL);

	return VID_SMALL|VID_MEDIUM|VID_LARGE|VID_GREYSCALE|VID_COLOR;
}
 
char *MeteorGrab_Attach(void)
{
	return ".grabControls.meteor";
}
 
void MeteorGrab_Detach(void)
{
}

/*ARGSUSED*/
grabproc_t *MeteorGrab_Start(int max_framerate, int config,
			 reconfigproc_t *reconfig)
{
int	tmp;
u_short status = 0;

	xmit_color = (config & VID_COLOR);
	if(ioctl(video, METEORSTATUS, &status) < 0) 
		fprintf(stderr, "Meteor grab: couldn't get status.\n");

	switch(status & (METEOR_STATUS_HCLK|METEOR_STATUS_FIDT)) {
	default:
		fprintf(stderr, "Meteor grab: No video signal found.\n");
		/* fall through */
	case METEOR_STATUS_FIDT:	/* NTSC */
		switch (config & VID_SIZEMASK) { 
		case VID_SMALL:
			geom.rows = NTSC_HEIGHT/2;
			geom.columns = NTSC_WIDTH/2;
			break;
		case VID_MEDIUM: 
			geom.rows = NTSC_HEIGHT;
			geom.columns = NTSC_WIDTH;
			break;
		case VID_LARGE:
			geom.rows = NTSC_HEIGHT*2;
			geom.columns = NTSC_WIDTH*2;
			break;
		}
		break;
	case 0:				/* PAL */
		switch (config & VID_SIZEMASK) { 
		case VID_SMALL:
			geom.rows = PAL_HEIGHT/2;
			geom.columns = PAL_WIDTH/2;
			break;
		case VID_MEDIUM: 
			geom.rows = PAL_HEIGHT;
			geom.columns = PAL_WIDTH;
			break;
		case VID_LARGE:
			geom.rows = PAL_HEIGHT*2;
			geom.columns = PAL_WIDTH*2;
			break;
		}
		break;
	}
 
        if (ioctl(video, METEORSETGEO, &geom) < 0)
		return(0);

#ifndef METEOR_SINGLE
	tmp = METEOR_CAP_SINGLE;
	ioctl(video, METEORCAPTUR, &tmp);
	tmp = METEOR_CAP_CONTINOUS;
	ioctl(video, METEORCAPTUR, &tmp);
#endif

	(*reconfig)(xmit_color, geom.columns, geom.rows);


	return MeteorGrab_GetFrame;

}

void MeteorGrab_Stop(void)
{
#ifndef METEOR_SINGLE
int	tmp;

	tmp = METEOR_CAP_STOP_CONT ;
        ioctl(video, METEORCAPTUR, &tmp);
#endif
	/*close(video);*/
}

#endif /* MeteorGRAB */

